{ File   : SFADECOD.PAS
  Author : Deniz Oezmen
  Created: ~2000
  Changed: 2005-08-25

  Star Trek: Starfleet Academy DAT extractor.
}
program SFADecode;
{$M 65520, 0, 655360}
uses
  TypeChg2,
  Crt;

var
  SourceF,
  DestF  : file;
  Log    : Text;

const
  UncompressedFormat = 32;
  CompressedFormat   = 64;
  LogName            = 'SFADECOD.LOG';

type
  TDATHeader = record
                 NrDirectories,
                 Unknown1,
                 Unknown2,
                 Unknown3   : Longint
               end;

  TFileEntry = record
                 ID,              (* 32 : unkomprimiert, 64: komprimiert *)
                 FOfs,            (* absolutes Dateistartoffset *)
                 FSize,           (* unkomprimierte Dateigre *)
                 FCompSize : Longint (* komprimierte Dateigre *)
               end;

  TBlockHeader = record
                   NrEntries,
                   Unknown1,
                   Unknown2,
                   Unknown3 : Longint
                 end;
  TElement = string;
  {$I LISTE.INC}

var
  DATHeader: TDATHeader;
  DirList  : TListe;

function ReadBE: Longint;
var
  Temp: Longint;
  B   : Byte;
begin
  Temp := 0;
  BlockRead(SourceF, B, 1);
  Inc(Temp, Longint(B) * 16777216);
  BlockRead(SourceF, B, 1);
  Inc(Temp, Longint(B) * 65536);
  BlockRead(SourceF, B, 1);
  Inc(Temp, Longint(B) * 256);
  BlockRead(SourceF, B, 1);
  Inc(Temp, Longint(B));
  ReadBE := Temp
end;

function ReadWordBE: Longint;
var
  Temp: Longint;
  B   : Byte;
begin
  Temp := 0;
  BlockRead(SourceF, B, 1);
  Inc(Temp, Longint(B) * 256);
  BlockRead(SourceF, B, 1);
  Inc(Temp, Longint(B));
  ReadWordBE := Temp
end;

procedure MakeDirTree(DirName: string);
var
  NextPos,
  LastPos: Longint;
  NextDir: string;
begin
  LastPos := 0;
  NextDir := '';
  while Length(NextDir) < Length(DirName) do
  begin
    (* Nchste Ebene bestimmen *)
    NextPos := LastPos + Pos('\', Copy(DirName, LastPos + 1,
      Length(DirName) - LastPos));
    if NextPos <> LastPos then
      NextDir := NextDir + Copy(DirName, LastPos + 1, NextPos - LastPos)
    else
      NextDir := NextDir + Copy(DirName, LastPos + 1,
        Length(DirName) - LastPos);
    LastPos := NextPos;
    (* Directory erstellen *)
    {$I-}
    if Copy(NextDir, Length(NextDir), 1) <> '\' then
      MkDir(NextDir)
    else
      MkDir(Copy(NextDir, 1, Length(NextDir) - 1))
    {$I+}
  end;
  if IOResult <> 0 then
    ;
end;

function ReadHeader: Boolean;
var
  Temp: Word;
begin
  with DATHeader do
  begin
    NrDirectories := ReadBE;
    Unknown1      := ReadBE;
    Unknown2      := ReadBE;
    Unknown3      := ReadBE
  end;
  Writeln('  Number of directory entries: ', DATHeader.NrDirectories);
  Writeln(Log, '  Number of directory entries: ', DATHeader.NrDirectories);
  Writeln(Log, '  HeaderInfo1: ', DATHeader.Unknown1);
  Writeln(Log, '  HeaderInfo2: ', DATHeader.Unknown2);
  Writeln(Log, '  HeaderInfo3: ', DATHeader.Unknown3);
  ReadHeader := True
end;

procedure ListDirectories;
var
  i, j,
  DirNameSize: Longint;
  DirName    : string;
  TempChar   : Char;
begin
  for i := 1 to DATHeader.NrDirectories do
  begin
    DirNameSize := 0;
    DirName     := '';
    BlockRead(SourceF, DirNameSize, 1);
    for j := 1 to DirNameSize do
    begin
      BlockRead(SourceF, TempChar, 1);
      DirName := DirName + TempChar
    end;
    Writeln('Creating directory ', DirName);
    Writeln(Log, 'Creating directory ', DirName);
    if DirName <> '.' then MakeDirTree(DirName);
    if l_empty(DirList) then l_insert(DirList, DirName)
    else l_last(DirList, DirName)
  end
end;

procedure ListFiles;
const
  N = 4096;
  F = 18;
  THRESHOLD = 2;
var
  BufWrtPtr  : Integer;
  b,
  FlagByte   : Byte;
  i, j, k, l,
  m,
  ReadCount,
  BlockCount,
  BlockSize,
  OldPos,
  Offset,
  Length,
  FNameSize  : Longint;
  DirName,
  FName      : string;
  Buf        : array[1..20480] of Char;
  HisBuf     : array[0..(N + F - 1) - 1] of Byte;
  FileEntry  : TFileEntry;
  BlockHeader: TBlockHeader;
  RevMode    : Boolean;
begin
  l_reset(DirList);
  for i := 1 to DATHeader.NrDirectories do
  begin
    with BlockHeader do
    begin
      NrEntries := ReadBE;
      Unknown1  := ReadBE;
      Unknown2  := ReadBE;
      Unknown3  := ReadBE
    end;
    Writeln(Log, '');
    Writeln(Log, 'Starting next directory...');
    Writeln(Log, '  # of files in this directory: ', BlockHeader.NrEntries);
    Writeln(Log, '  block info 1                : ', BlockHeader.Unknown1);
    Writeln(Log, '  block info 2                : ', BlockHeader.Unknown2);
    Writeln(Log, '  block info 3                : ', BlockHeader.Unknown3);
    l_show(DirList, DirName);
    l_next(DirList);
    for j := 1 to BlockHeader.NrEntries do
    begin
      FNameSize := 0;
      FName := '';
      BlockRead(SourceF, FNameSize, 1);
      for k := 1 to FNameSize do
      begin
        BlockRead(SourceF, b, 1);
        FName := FName + Chr(b)
      end;
      with FileEntry do
      begin
        ID        := ReadBE;
        FOfs      := ReadBE;
        FSize     := ReadBE;
        FCompSize := ReadBE;
      end;
      Write('Extracting ', DirName + '\' + FName, '... ');
      Write(Log, 'Extracting ', DirName + '\' + FName, '... ');
      OldPos := FilePos(SourceF);
      if (FileEntry.ID = UncompressedFormat) then
      begin
        Assign(DestF, DirName + '\' + FName);
        Rewrite(DestF, 1);
        Seek(SourceF, FileEntry.FOfs);
        for l := 1 to FileEntry.FSize div SizeOf(Buf) do
        begin
          BlockRead(SourceF, Buf, SizeOf(Buf));
          BlockWrite(DestF, Buf, SizeOf(Buf))
        end;
        BlockRead(SourceF, Buf, FileEntry.FSize mod SizeOf(Buf));
        BlockWrite(DestF, Buf, FileEntry.FSize mod SizeOF(Buf));
        Close(DestF);
        Writeln('finished');
        Writeln(Log, 'finished');
        Writeln(Log, '  File ID    : ', FileEntry.ID);
        Writeln(Log, '  File Offset: ', FileEntry.FOfs);
        Writeln(Log, '  File size  : ', FileEntry.FSize);
        Writeln(Log, '  Unused     : ', FileEntry.FCompSize)
      end
      else if FileEntry.ID = CompressedFormat then
      begin
        Assign(DestF, DirName + '\' + FName);
        Rewrite(DestF, 1);
        Writeln('--> compressed file (',
          FileEntry.FCompSize / FileEntry.FSize * 100 : 5 : 2, '%)');
        Write('  Decompressing... ');
        Writeln(Log, '--> compressed file (',
          FileEntry.FCompSize / FileEntry.FSize * 100 : 5 : 2, '%)');
        Writeln(Log, '  File ID          : ', FileEntry.ID);
        Writeln(Log, '  File Offset      : ', FileEntry.FOfs);
        Writeln(Log, '  uncompressed size: ', FileEntry.FSize);
        Writeln(Log, '  compressed size  : ', FileEntry.FCompSize);
        Seek(SourceF, FileEntry.FOfs);
        ReadCount := 0;
        while ReadCount < FileEntry.FCompSize do
        begin
          BlockCount := 0;
          BlockSize := ReadWordBE;
          RevMode := (BlockSize and $8000) > 0;
          BlockSize := BlockSize and $7FFF;
          Inc(ReadCount, 2);
          if RevMode then
          begin
            for l := 1 to BlockSize div SizeOf(Buf) do
            begin
              BlockRead(SourceF, Buf, SizeOf(Buf));
              BlockWrite(DestF, Buf, SizeOf(Buf))
            end;
            BlockRead(SourceF, Buf, BlockSize mod SizeOf(Buf));
            BlockWrite(DestF, Buf, BlockSize mod SizeOF(Buf));
            BlockCount := BlockSize
          end
          else
          begin
            BufWrtPtr := Longint(N - F);
            for l := 0 to sizeof(HisBuf) - 1 do
              HisBuf[l] := 32;
            while (BlockCount < BlockSize) do
            begin
              BlockRead(SourceF, FlagByte, 1);
              Inc(BlockCount);
              for l := 1 to 8 do
              begin
                if (FlagByte and $01 = 0) and
                  (BlockCount < BlockSize - 1) then
                begin
                  Offset := 0;
                  Length := 0;
                  BlockRead(SourceF, Offset, 1);
                  BlockRead(SourceF, Length, 1);
                  Inc(BlockCount, 2);
                  Offset := Offset or ((Length and $F0) shl 4);
                  Length := (Length and $0F) + THRESHOLD;
                  for m := 0 to Length do
                  begin
                    b := HisBuf[Longint(N - 1) and Longint(Offset + m)];
                    BlockWrite(DestF, b, 1);
                    HisBuf[BufWrtPtr] := b;
                    Inc(BufWrtPtr);
                    BufWrtPtr := Longint(N - 1) and Longint(BufWrtPtr)
                  end
                end
                else if BlockCount < BlockSize then
                begin
                  BlockRead(SourceF, b, 1);
                  Inc(BlockCount);
                  BlockWrite(DestF, b, 1);
                  HisBuf[BufWrtPtr] := b;
                  Inc(BufWrtPtr);
                  BufWrtPtr := Longint(N - 1) and Longint(BufWrtPtr);
                end;
                FlagByte := FlagByte shr 1
              end
            end
          end;
          Inc(ReadCount, BlockCount)
        end;
        Writeln(Log, '  Bytes written    : ', FileSize(DestF));
        Writeln(Log, '  Bytes read       : ', ReadCount);
        Close(DestF);
        Writeln('finished')
      end
      else
      begin
        Writeln(' --> unknown file entry format!');
        Writeln(Log, ' --> unknown file entry format!');
        Writeln(Log, '  File ID was ', FileEntry.ID)
      end;
      Seek(SourceF, OldPos)
    end
  end
end;

function ParseCommandLine: Boolean;
begin
  ParseCommandLine := (ParamCount = 1)
end;

function OpenSource : Boolean;
begin
  if ParamStr(1) <> '' then
  begin
    Assign(SourceF, ParamStr(1));
    {$I-}
    Reset(SourceF, 1);
    {$I+}
    OpenSource := (IOResult = 0)
  end
  else
    OpenSource := FALSE
end;

procedure AnalyseCommandLine;
begin
  {maybe somewhen switches to filter out file types}
  Writeln('  brute-force extraction mode set');
  Writeln(Log, '  brute-force extraction mode set')
end;

procedure OpenLog;
begin
  Assign(Log, LogName);
  Rewrite(Log)
end;

procedure CloseLog;
begin
  Close(Log)
end;

begin
  Writeln('SFADECOD -- Starfleet Academy source file extractor');
  Writeln('(C)opyright Deniz Oezmen, 2000');
  Writeln('early alpha - use at your own risk');
  Writeln;
  OpenLog;
  Writeln(Log, 'Log file for SFADECOD');
  Writeln(Log, 'SFADECOD is (C) Deniz Oezmen, 2000');
  Writeln(Log, '');
  Write(' Checking command line... ');
  Write(Log, ' Checking command line... ');
  if ParseCommandLine then
  begin
    Writeln('OK');
    Writeln(Log, 'OK');
    AnalyseCommandLine;
    Write(' Opening ', ParamStr(1), '... ');
    Write(Log, ' Opening ', ParamStr(1), '... ');
    if OpenSource then
    begin
      Writeln('successful');
      Writeln(Log, 'successful');
      Writeln(' Reading file header... ');
      Writeln(Log, ' Reading file header... ');
      if ReadHeader then
      begin
        Writeln;
        Writeln(Log, '');
        Writeln(' Building directory structure...');
        Writeln(Log, ' Building directory structure...');
        l_create(DirList);
        ListDirectories;
        Writeln;
        Writeln(Log, '');
        Writeln(' Extracting files...');
        Writeln(Log, ' Extracting files...');
        ListFiles;
        Writeln;
        Writeln(Log, '');
        Write(' Clearing up and closing files... ');
        Write(Log, ' Clearing up and closing files... ');
        l_delete(DirList);
        Close(SourceF);
        Writeln('complete');
        Writeln(Log, 'complete')
      end
      else
      begin
        Writeln('unrecognized format');
        Writeln(Log, 'unrecognized format')
      end
    end
    else
    begin
      Writeln('failed');
      Writeln(Log, 'failed')
    end
  end
  else
  begin
    Writeln('incorrect number of parameters');
    Writeln(Log, 'incorrect number of parameters')
  end;
  CloseLog
end.
